'''Containing module for the Calculator class.'''

from typing import List, Dict, Union
from veq.tokens import TokenBuilder, Token, TokenStream, VariableToken, FunctionToken

class CalculationError(Exception):

    '''An exception that indicates there was an error in calculating the user's equation.'''
    
    def __init__(self, expression: str):
        self.__expression = expression
        super().__init__(expression)

    def __str__(self):
        return f"Error in equation \"f(x) = {self.__expression}\""

class Calculator:

    '''Class responsible for converting infix to postfix and calculating expressions.

    :attribute stream: Expression in infix format for reading expressions.
    :type stream: str or TokenStream
    :attribute stack: Calculation stack.
    :type stack: List(float)
    :value stack: []'''

    def __init__(self, stream: Union[str, TokenStream], stack: List[float] = []):
        self.stack = stack
        self.__builder = TokenBuilder(self.stack)

        if isinstance(stream, str):
            self.stream = TokenStream(stream)
        else:
            self.stream = stream

        self.expression = []

    def infix_to_postfix(self):
        '''Convert infix stream to postfix list of tokens.

        :param expression: String to tokenize.
        :type expression: str
        :returns: List of tokens
        :rtype: list(Token)'''

        lookup: Dict[str, Token] = {
            "+":   self.__builder.build_add,
            "-":   self.__builder.build_subtract,
            "*":   self.__builder.build_multiply,
            "/":   self.__builder.build_divide,
            "^":   self.__builder.build_power,
            "sin(": self.__builder.build_sin,
            "cos(": self.__builder.build_cos,
            "tan(": self.__builder.build_tan,
            "log(": self.__builder.build_log,
        }

        stack: List[Token] = []

        token: Token = None

        for match in self.stream:
            if match in lookup:
                token = lookup[match]()
            elif match == "x":
                token = self.__builder.build_x()
                self.expression.append(token)
                continue
            elif match == '(':
                self.infix_to_postfix()
                continue
            elif match == ')':
                while len(stack) > 0:
                    self.expression.append(stack.pop())
                return
            elif match == 'pi':
                token = self.__builder.build_pi()
                self.expression.append(token)
                continue
            elif match == 'e':
                token = self.__builder.build_e()
                self.expression.append(token)
                continue
            else:
                try:
                    x = float(match)
                except:
                    raise TypeError(f"Invalid Token \"{match}\"")
                else:
                    token = self.__builder.build_value(x)
                    self.expression.append(token)
                    continue

            if isinstance(token, FunctionToken):
                self.infix_to_postfix()
                self.expression.append(token)
                continue

            while len(stack) > 0 and stack[-1].precedence >= token.precedence:
                self.expression.append(stack.pop())
            stack.append(token)

        while len(stack) > 0:
            self.expression.append(stack.pop())

    def calculate(self, x: float) -> float:
        '''Calculate the Calculator's expression.

        :param x: Value to substitute for x in the expression.
        :type x: float
        :returns: The result of the equation.
        :rtype: float'''
        if len(self.expression) == 0:
            self.infix_to_postfix()

        self.stack.clear()

        try:
            for token in self.expression:
                if isinstance(token, VariableToken):
                    token.execute(x)
                else:
                    token.execute()
        except:
            pass
        else:
            if len(self.stack) > 0:
                result = self.stack[-1]
                if isinstance(result, (float, int)):
                    return result
        raise CalculationError(self.stream.text)



'''Module containing all tokens and classes needed to tokenize expressions.

:data TOKEN_REGEX: Regular Expression used for extracting specific token characters.
:type TOKEN_REGEX: Pattern'''

import re
from abc import ABC, abstractmethod
from typing import List, Pattern
from math import e, pi, sin, cos, tan, log

class Token(ABC):

    '''Token Abstract Base Class.'''

    def __init__(self, stack: List):
        self._stack = stack

    @abstractmethod
    def execute(self):
        '''Execute the function.'''
        pass

    @property
    def precedence(self):
        '''Precedence of a token.'''
        try:
            return self.__class__.PRECEDENCE
        except NameError: # If value isn't set
            return -1

class VariableToken(Token):

    '''Variable token.'''

    def execute(self, value: float):
        '''Push given value to top of stack.

        :param value: What value to push.
        :type value: float'''
        self._stack.append(value)

class ValueToken(Token):

    '''Token representing a number.'''

    def __init__(self, stack: List, value: float):
        super().__init__(stack)
        self.__value = value

    def execute(self):
        '''Push value to top of stack.'''
        self._stack.append(self.__value)

class PiToken(ValueToken):

    '''Token for the number pi.'''

    def __init__(self, stack: List):
        super().__init__(stack, pi)

class EulerToken(ValueToken):

    '''Token for the number e.'''

    def __init__(self, stack: List):
        super().__init__(stack, e)

class BinaryToken(Token):

    '''Binary Operation token.'''

    @abstractmethod
    def operation(self, a, b):
        '''What operation to execute.'''
        pass

    def execute(self):
        '''Pop two items from the stack, run an operation on them,\
            and push the result to the stack.'''
        b = self._stack.pop()
        a = self._stack.pop()
        result = self.operation(a, b)
        self._stack.append(result)

class AddToken(BinaryToken):

    '''Addition token.'''

    PRECEDENCE = 1

    def operation(self, a, b):
        '''Add a and b'''
        return a + b

class SubtractToken(BinaryToken):

    '''Subtraction token.'''

    PRECEDENCE = 1

    def operation(self, a, b):
        '''Subtract a and'''
        return a - b

class MultiplyToken(BinaryToken):

    '''Multiplication token.'''

    PRECEDENCE = 2

    def operation(self, a, b):
        '''Multiply a and b'''
        return a * b

class DivideToken(BinaryToken):

    '''Division token.'''

    PRECEDENCE = 2

    def operation(self, a, b):
        '''Divide a and b.'''
        return a / b

class PowerToken(BinaryToken):

    '''Exponation token.'''

    PRECEDENCE = 3

    def operation(self, a, b):
        '''Exponate a to the power of b'''
        return a ** b

class FunctionToken(Token):

    '''Token representing a unary function.'''

    @abstractmethod
    def operation(self, a):
        pass

    def execute(self):
        '''Pop an item from the stack, run a function on it,\
            and push the result to the stack.'''
        a = self._stack.pop()
        result = self.operation(a)
        self._stack.append(result)

class SinToken(FunctionToken):

    '''Sine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run sin'''
        return sin(a)

class CosToken(FunctionToken):

    '''Cosine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run cosine'''
        return cos(a)

class TanToken(FunctionToken):

    '''Tangent token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run tangent'''
        return tan(a)

class LogToken(FunctionToken):

    '''Logarithm token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run logarithm base 10.'''
        return log(a)

class TokenBuilder:

    '''Build tokens.'''

    def __init__(self, stack: List):
        self._stack = stack

    def build_value(self, value: float):
        '''Build a value token.'''
        return ValueToken(self._stack, value)

    def build_add(self):
        '''Build an add token.'''
        return AddToken(self._stack)

    def build_subtract(self):
        '''Build a subtract token.'''
        return SubtractToken(self._stack)

    def build_multiply(self):
        '''Build a multiply token.'''
        return MultiplyToken(self._stack)

    def build_divide(self):
        '''Build a divide token.'''
        return DivideToken(self._stack)

    def build_power(self):
        '''Build a power token.'''
        return PowerToken(self._stack)

    def build_pi(self):
        '''Build a pi token.'''
        return PiToken(self._stack)

    def build_e(self):
        '''Build an e token.'''
        return EulerToken(self._stack)

    def build_sin(self):
        '''Build a sin token.'''
        return SinToken(self._stack)

    def build_cos(self):
        '''Build a cos token.'''
        return CosToken(self._stack)

    def build_tan(self):
        '''Build a tan token.'''
        return TanToken(self._stack)

    def build_log(self):
        '''Build a log token.'''
        return LogToken(self._stack)

    def build_x(self):
        '''Build a variable token.'''
        return VariableToken(self._stack)

TOKEN_REGEX:Pattern[str] = re.compile(r"\d+\.?\d*|\+|-|\*|/|\)|\^|tan\(|cos\(|sin\(|log\(|pi|e|x|\(")

class TokenStream:

    '''A read-only representation of a stream of strings representing individual tokens.

    :property text: The infix represenetation of the full expression.
    :type text: str'''

    def __init__(self, expression: str):
        self.__expression: List[str] = TOKEN_REGEX.findall(expression)
        self.__iteration = 0

    @property
    def text(self) -> str:
        '''The infix representation of the full expression.'''
        return "".join(self.__expression)

    def reset(self):
        '''Reset the stream to the beginning.'''
        self.__iteration = 0

    def __iter__(self): 
        return self

    def __next__(self):
        if self.__iteration < len(self.__expression):
            value = self.__expression[self.__iteration]
            self.__iteration += 1
            return value
        raise StopIteration

    def __str__(self):
        return self.text



'''Main module.'''

import argparse
import logging
import pygame
from veq.calculator import Calculator
from veq.visualizer import Equation, Visualizer

def parse_args() -> argparse.Namespace:
    '''Parse command line arguments.

    :returns: Parsed arguments.
    :rtype; argparse.Namespace'''
    parser = argparse.ArgumentParser("veq", description="Visualize an equation.")
    parser.add_argument("equation", type=str.lower, help="Equation to plot.")
    parser.add_argument("-d", "--debug", dest="level", action="store_const",\
         const=logging.DEBUG, default=logging.INFO, help="Enable debug logging.")
    parser.add_argument("-s", "--step", metavar='n', type=float, default=None,\
         help="Enable gridlines with step n")
    parser.add_argument("-a", "--noaxis", dest="axis", action="store_false", help="Disable axis.")

    return parser.parse_args()

def main():
    '''Run the main functionality of the program.'''
    args = parse_args()

    logging.basicConfig(level=args.level)

    domain = [-1, 1]
    range_ = [-1, 1]

    calculator = Calculator(args.equation)

    pygame.init()

    screen = pygame.display.set_mode([900, 900])
    pygame.display.set_caption('veq')

    equation = Equation(calculator, domain.copy(), range_.copy())
    visualizer = Visualizer(equation, screen)

    running = True
    while running: 
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.MOUSEWHEEL:
                if event.y != 0:
                    equation.zoom(-event.y)
            elif event.type == pygame.MOUSEMOTION:
                mouse_x, mouse_y = event.pos

                # If RMB is down...
                if event.buttons[2] == 1:
                    rel = [x/100 for x in event.rel]
                    equation.domain[0] -= rel[0]
                    equation.domain[1] -= rel[0]
                    equation.range[0] += rel[1]
                    equation.range[1] += rel[1]
            elif event.type == pygame.MOUSEBUTTONUP:
                if event.button == 1:
                    visualizer.save(event.pos)
            elif event.type == pygame.KEYUP:
                logging.debug("Key pressed: %i", event.key)
                if event.key == pygame.K_MINUS or event.key == pygame.K_KP_MINUS:
                    equation.zoom(1)
                elif event.key == pygame.K_PLUS or event.key == 61 or event.key == pygame.K_KP_PLUS:
                    # 61 = Plus. For some reason pygame.K_PLUS wasn't wanting to work
                    equation.zoom(-1)
                elif event.key == pygame.K_r:
                    equation.domain = domain.copy()
                    equation.range = range_.copy()
            elif event.type == pygame.ACTIVEEVENT:
                focus = (event.gain == 1)

        screen.fill((255, 255, 255))
        if not args.step is None:
            visualizer.draw_grid(args.step)
        if args.axis:
            visualizer.draw_axis()
        visualizer.draw_equation()
        visualizer.draw_text()

        if focus:
            visualizer.draw_location((mouse_x, mouse_y))
        visualizer.draw_saved()

        pygame.display.flip()

    pygame.quit()



if __name__ == "__main__":
    main()
